/*	$NetBSD: usb_mem.h,v 1.23 2005/12/11 12:24:01 christos Exp $	*/
/*	$FreeBSD: src/sys/dev/usb/usb_mem.h,v 1.21 2005/01/06 01:43:29 imp Exp $	*/

/*-
 * Copyright (c) 1998, 2007 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Lennart Augustsson (lennart@augustsson.net) at
 * Carlstedt Research & Technology.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by ITOH Yasufumi (itohy@NetBSD.org).
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#define USB_DMA_NSEG (btoc(MAXPHYS) + 1)

typedef struct usb_dma_block {
#ifdef __FreeBSD__
	bus_dma_tag_t tag;
#endif
	bus_dmamap_t map;
        caddr_t kaddr;
        bus_dma_segment_t segs[1];
        int nsegs;
        size_t size;
        size_t align;
	int flags;
#define USB_DMA_FULLBLOCK	0x0001
#define USB_DMA_MARKED_TO_FREE	0x0002
	LIST_ENTRY(usb_dma_block) next;
} usb_dma_block_t;

#if defined(__NetBSD__) || defined(__OpenBSD__)
#define DMAADDR(dma, o) \
	((dma)->block->map->dm_segs[0].ds_addr + (dma)->offs + (o))
#define USB_MEM_SYNC(utag, dma, ops) \
	bus_dmamap_sync((utag)->tag, (dma)->block->map, \
			(dma)->offs, (dma)->len, (ops))
#define USB_MEM_SYNC2(utag, dma, off, len, ops) \
	bus_dmamap_sync((utag)->tag, (dma)->block->map, \
			(dma)->offs + (off), (len), (ops))
#elif defined(__FreeBSD__)
#define DMAADDR(dma, o) \
	((dma)->block->segs[0].ds_addr + (dma)->offs + (o))
/*
 * XXX overkill
 * FreeBSD can't sync map partially
 */
#define USB_MEM_SYNC(utag, dma, ops) \
	bus_dmamap_sync((dma)->block->tag, (dma)->block->map, (ops))
#define USB_MEM_SYNC2(utag, dma, off, len, ops) \
	bus_dmamap_sync((dma)->block->tag, (dma)->block->map, (ops))
#endif

#define KERNADDR(dma, o) \
	((void *)((dma)->block->kaddr + (dma)->offs + (o)))

typedef struct {
	struct usb_dma_block *block;
	u_int offs;
	u_int len;
} usb_dma_t;

/*
 * DMA buffer
 */
struct usb_buffer_dma {
	enum { UB_NONE, UB_BUF, UB_ALLOC_MAP, UB_MAP } ub_type;
	usb_dma_t	ub_allocbuf;
#if defined(__NetBSD__) || defined(__OpenBSD__)
	bus_dma_segment_t ub_segs[USB_DMA_NSEG];
	int		ub_nsegs;
	size_t		ub_len;
	bus_dmamap_t	ub_map;
	caddr_t		ub_kaddr;
#define _USB_BUFFER_NSEGS(ub)	(ub)->ub_map->dm_nsegs
#define USB_BUFFER_NSEGS(ub)	\
	((ub)->ub_type == UB_BUF ? 1 : _USB_BUFFER_NSEGS(ub))
#define USB_BUFFER_SEGS(ub)	\
	((ub)->ub_type == UB_BUF ? \
	 (ub)->ub_segs /* XXX abuse */ : (ub)->ub_map->dm_segs)
#elif defined(__FreeBSD__)
	bus_dma_tag_t	ub_tag;
	void		*ub_kaddr;
	bus_dma_segment_t ub_segs[USB_DMA_NSEG];
	int		ub_nsegs;
	bus_dmamap_t	ub_map;
	int		ub_error;
#define _USB_BUFFER_NSEGS(ub)	(ub)->ub_nsegs
#define USB_BUFFER_NSEGS(ub)	(_USB_BUFFER_NSEGS(ub))
#define USB_BUFFER_SEGS(ub)	(ub)->ub_segs
#endif
	SIMPLEQ_ENTRY(usb_buffer_dma)	ub_next;
};

/*
 * per host controller
 */
typedef struct usb_dma_tag {
	/*
	 * bus_dma(9) tag, filled by host controller driver
	 */
#if defined(__NetBSD__) || defined(__OpenBSD__)
	bus_dma_tag_t	tag;
#endif
#ifdef __FreeBSD__
	bus_dma_tag_t	tag_parent;
	bus_dma_tag_t	tag_buffer;
#endif

	/*
	 * Lists of free DMA memory
	 */
	LIST_HEAD(, usb_dma_block) blk_freelist;
	int	blk_nfree;
	LIST_HEAD(, usb_frag_dma) frag_freelist;
	/* for statistical and diagnostic purposes */
	int	nblks;		/* number of allocated buffers */
	int	nfrags;		/* number of allocated fragments */

	/* for delayed free */
	SIMPLEQ_HEAD(, usb_buffer_dma) ubs_idle, ubs_tobefreed;
	/* for statistical and diagnostic purposes */
	int	nbufs;		/* number of allocated buffer structures */
} usb_dma_tag_t;

usbd_status	usb_allocmem(usb_dma_tag_t *, size_t, size_t, usb_dma_t *);
void		usb_freemem(usb_dma_tag_t *, usb_dma_t *);
usbd_status	usb_alloc_buffer_dma(usb_dma_tag_t *, void *, size_t,
		    struct usb_buffer_dma **, void **);
void		usb_free_buffer_dma(usb_dma_tag_t *, struct usb_buffer_dma *,
		    enum usbd_waitflg);
void		usb_sync_buffer_dma(usb_dma_tag_t *, struct usb_buffer_dma *,
		    int /*op*/);
usbd_status	usb_insert_transfer_dma(usbd_xfer_handle, usb_dma_tag_t *,
		    struct usb_buffer_dma *);
void		usb_transfer_complete_dma(usbd_xfer_handle, usb_dma_tag_t *,
		    struct usb_buffer_dma *);
void		usb_dma_tag_init(usb_dma_tag_t *);
void		usb_dma_tag_finish(usb_dma_tag_t *);
